#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include "get_stat.h"
#include "parse_date.h"

int main(int argc, char **argv) {
	time_t t = time(NULL);
	char *fmt = "%a %b %e %H:%M:%S %Z %Y";

	/* For -s/-r flag */
	struct timespec ts;

	char *r_flag = NULL;

	int opt;
	while ((opt = getopt(argc, argv, "s:d:ur:")) != -1) {
		switch (opt) {
			case 's':
				ts.tv_sec = mu_parse_date("date", optarg);
				if (ts.tv_sec == -1)
					return 1;

				if (clock_settime(CLOCK_REALTIME, &ts) < 0) {
					fprintf(stderr, "date: settime: %s\n", strerror(errno));
					return 1;
				}

				return 0;

			case 'd':
				t = mu_parse_date("date", optarg);
				if (t == -1)
					return 1;

				break;

			case 'u':
				if (setenv("TZ", "UTC0", 1) < 0) {
					fprintf(stderr, "date: %s\n", strerror(errno));
					return 1;
				}

				break;

			case 'r':
				r_flag = optarg;
				break;

			default:
				puts("date [sdur] [+\"fmt\"]\n\t-s TIME Set new date\n\t-d TIME Print new date\n\t-u Work in UTC\n\t-r FILE Display last modification time of FILE\n\nRecognized TIME formats:");
				for (size_t i = 0; i < sizeof(mu_date_fmts) / sizeof(char *); i++)
					puts(mu_date_fmts[i]);

				return 0;
		}
	}

	argv += optind;
	argc -= optind;

	if (argc != 0)
		if (argv[0][0] == '+')
			fmt = argv[0] + 1;

	if (r_flag) {
		struct stat sb;
		if (mu_get_lstat("date", r_flag, &sb))
			return 1;

		t = sb.st_mtime;
	}

	struct tm *tm = localtime(&t);

	char buf[256];
	strftime(buf, sizeof(buf), fmt, tm);

	puts(buf);
	return 0;
}
